import serial
from serial.tools import list_ports

import threading
import time
import datetime

from dynamikontrol.Protocol import Module2PC, PC2Module
from dynamikontrol.LED import LED
from dynamikontrol.BaseLED import BaseLED
from dynamikontrol.Motor import Motor
from dynamikontrol.helpers.helper import print_bytearray

class Module(object):
    """Module class.

    .. highlight:: python
    .. code-block:: python

        from dynamikontrol import Module

        module = Module(serial_no) # specify the module by serial number

        # Print module serial number
        print('Serial number: %s' % (module.get_serial_no(),)

        module.disconnect()

    Args:
        serial_no (str): Specify serial number of the module.
        debug (bool): print debug messages.
    """
    __serial_receive_delay = 0
    __receive_thread_delay = 0.001
    __stop_thread = False

    # communication
    __is_header_defined = False

    def __init__(self, serial_no=None, debug=False):
        self.serial_no = serial_no
        if self.serial_no is not None:
            self.serial_no = self.serial_no.lower()
        self.debug = debug
        self.p2m = PC2Module()
        self.m2p = Module2PC()

        self.fw_type = 0
        self.fw_version = None

        self.port = None
        self.ser = None
        self.baud = 115200
        self.vid = None
        self.pid = None
        self.avail_vids = ['3476']
        self.avail_pids = ['0001', '0002'] # [Angle, Speed]
        self.serial_no_len = 8
        self.id_len = 1
        self.time_len = 8
        self.fw_version_len = 8
        self.manual_delay = 0.1
        self.data_queue = bytearray()

        self.is_connected = False

        ports = list_ports.comports(include_links=True)

        for port in ports:
            if self.debug:
                for i, p in enumerate(port):
                    print('[*] port[%d] %s' % (i, p))

            desc = port[2].lower()
            if 'vid:pid' not in desc:
                continue

            start_idx = desc.find('vid:pid=')
            self.vid, self.pid = desc[start_idx+8:start_idx+8+9].split(':')

            if self.vid not in self.avail_vids or self.pid not in self.avail_pids:
                continue

            start_idx = desc.find('ser=')
            serial_no_str = desc[start_idx+4:start_idx+4+self.serial_no_len]

            if self.serial_no is not None and self.serial_no != serial_no_str:
                continue

            self.port = port[0]
            break

        self.connect()

        self.led = LED(module=self)
        self.base_led = BaseLED(module=self)
        self.motor = Motor(module=self)


    def connect(self):
        """Connect to the module.
        """
        if self.port is None:
            raise IOError('Module is not connected!')

        self.ser = serial.Serial(self.port, self.baud, timeout=0)

        if self.debug:
            print('[*] Connected to %s, baud rate: %d' % (self.port, self.baud))

        self.get_fw_version()
        time.sleep(0.1)

        self.receive_thread = threading.Thread(target=self.__receive, args=())
        self.receive_thread._event = threading.Event()
        self.receive_thread.daemon = True
        self.receive_thread.start()
        self.receive_thread._event.set()

        self.send(self.p2m.set_type(0x00).set_command(0x00).set_data([]).encode()) # connect
        time.sleep(self.manual_delay)


    def disconnect(self):
        """Close the connection of the module. Must include ``module.disconnect()`` at the end of the code so that module can close connection properly.
        """
        time.sleep(1)
        if self.debug:
            print('[*] Disconnecting...')
        self.__stop_thread = True
        time.sleep(1)
        self.ser.close()


    def __read_delay(self, size=1):
        """Read the data from the module using serial communication.

        Args:
            size (int, optional): Length of bytes. Defaults to ``1``.

        Returns:
            int: Data from the module.
        """
        time.sleep(self.__serial_receive_delay)
        return int.from_bytes(self.ser.read(size), byteorder='little')


    def __receive(self):
        """Receive data from the module. It runs on single thread for waiting response from the module.
        """
        while not self.is_connected:
            self.is_connected = True
            self.__stop_thread = False

            while not self.__stop_thread:
                self.receive_thread._event.wait()

                time.sleep(self.__receive_thread_delay)

                try:
                    header = self.__read_delay()

                    if not header:
                        continue

                    if not self.__is_header_defined and (header == 0x06 or header == 0x15): # 0x06 ACK, 0x15 NACK
                        self.__is_header_defined = True
                        self.data_queue = bytearray()

                    if not self.__is_header_defined:
                        continue

                    # data queue
                    self.data_queue.append(header) # header
                    self.data_queue.append(self.__read_delay()) # type
                    self.data_queue.append(self.__read_delay()) # command

                    # data_length
                    data_length = self.__read_delay()
                    self.data_queue.append(data_length)

                    # data
                    for i in range(data_length):
                        self.data_queue.append(self.__read_delay())

                    self.data_queue.append(self.__read_delay()) # checksum

                    # end
                    end = self.__read_delay()
                    self.data_queue.append(end)

                    if end != 0x04:
                        self.__is_header_defined = False
                        raise ValueError('Module invalid end byte')

                    # command, data = self.m2p.decode(self.data_queue)

                    if self.debug:
                        print('[*] Recv %s' % (print_bytearray(self.data_queue),))
                except Exception as e:
                    print(e)
                finally:
                    self.__is_header_defined = False

        self.is_connected = False
        self.__stop_thread = False


    def _manual_send_receive(self, send_data, receive_data_len):
        """Pause receiving thread, send/receive data manually and resume receiving thread.

        Args:
            send_data (bytearray of int): Data to send.
            receive_data_len (int): Length of the data to receive.

        Returns:
            tuple: (command, received data)
        """
        if hasattr(self, 'receive_thread'):
            self.receive_thread._event.clear() # pause receive thread

        self.send(send_data)
        time.sleep(self.manual_delay)

        received_data = bytearray()

        for i in range(receive_data_len):
            received_data.append(self.__read_delay())

        command, received_data = self.m2p.decode(received_data)

        if hasattr(self, 'receive_thread'):
            self.receive_thread._event.set() # resume receive thread

        return command, received_data


    def send(self, data):
        """Send the data to the module

        Args:
            data (bytearray of int): Data to send.
        """
        if self.ser is None:
            raise IOError('Serial is not connected!')

        if self.port is None:
            raise IOError('Module is not connected!')

        if type(data) == str:
            self.ser.write(data.encode())
        else:
            self.ser.write(data)

        if self.debug:
            print('[*] Sent %s' % (print_bytearray(data),))


    def get_serial_no(self):
        """Get serial number of the connected module.

        Returns:
            str: Serial number.
        """
        command, serial_no = self._manual_send_receive(
            self.p2m.set_type(0x00).set_command(0x80).set_data([]).encode(),
            self.serial_no_len + 6
        )
        return serial_no.decode('utf-8')


    def get_id(self):
        """Get ID of the connected module.

        Returns:
            int: Module ID.
        """
        command, id = self._manual_send_receive(
            self.p2m.set_type(0x00).set_command(0x81).set_data([]).encode(),
            self.id_len + 6
        )
        return id[0]


    def get_time(self):
        """Get device time of the connected module.

        Returns:
            datetime: Device time.
        """
        command, device_time = self._manual_send_receive(
            self.p2m.set_type(0x00).set_command(0x82).set_data([]).encode(),
            self.time_len + 6
        )

        # TODO: get_time() year, month, day
        h, m, s, ms, _, _, _, _ = device_time

        now = datetime.datetime.now()

        return datetime.datetime(now.year, now.month, now.day, h, m, s, ms)


    def get_fw_version(self):
        """Get firmware version of the connected module.

        Returns:
            str: Device firmware version.
        """
        command, fw_data = self._manual_send_receive(
            self.p2m.set_type(0x00).set_command(0x83).set_data([]).encode(),
            self.fw_version_len + 6
        )

        self.fw_type = fw_data[0]

        fw_major_version = int(fw_data[1:4].decode('utf-8'))
        fw_minor_version = fw_data[5:8].decode('utf-8')
        self.fw_version = '%s.%s' % (fw_major_version, fw_minor_version)

        return self.fw_version



import time
from datetime import datetime
import threading

class Timer(object):
    """General timer class.

    .. highlight:: python
    .. code-block:: python

        from dynamikontrol import Module, Timer
        import time

        t1 = Timer()
        t2 = Timer()

        module = Module()

        t1.callback_at(func=module.led.toggle, args=('r',), at='2021-03-02 19:46:30', interval=0.1)

        t2.callback_after(func=module.led.toggle, args=('g',), after=1, interval=0.1)

        time.sleep(5)

        t1.stop()
        t2.stop()

        module.disconnect()

    """
    __stop_thread = False

    def __init__(self):
        pass

    def callback_after(self, func, args=(), kwargs={}, after=0, interval=None):
        """Call the callback function after specific time.

        Args:
            func (function): Callback function.
            args (tuple, optional): args. Defaults to ``()``.
            kwargs (dict, optional): kwargs. Defaults to ``{}``.
            after (int, optional): Callback delay time in seconds. Defaults to ``0``.
            interval (int, optional): Callback interval time in seconds. Defaults to ``None``.
        """
        self.__stop_thread = False
        if interval is None:
            self.__stop_thread = True

        def handler():
            time.sleep(after)
            func(*args, **kwargs)

            next_call = time.time()
            while True:
                if self.__stop_thread:
                    break

                next_call = next_call + interval
                time.sleep(max(next_call - time.time(), 0))
                func(*args, **kwargs)

        timer_thread = threading.Thread(target=handler)
        timer_thread.start()

    def callback_at(self, func, args=(), kwargs={}, at=None, interval=None):
        """Call the callback function at specific time.

        Args:
            func (function): Callback function.
            args (tuple, optional): args. Defaults to ``()``.
            kwargs (dict, optional): kwargs. Defaults to ``{}``.
            at (datetime str, optional): Callback time in datetime str. e.g) ``2021-03-04 21:57:30``. Defaults to ``None``.
            interval ([type], optional): Callback interval time in seconds. Defaults to ``None``.
        """
        after = datetime.strptime(at, '%Y-%m-%d %H:%M:%S').timestamp() - datetime.now().timestamp()

        if after < 0:
            return False

        self.callback_after(func=func, args=args, kwargs=kwargs, after=after, interval=interval)

    def stop(self):
        """Stop the timer.
        """
        self.__stop_thread = True



class Protocol(object):
    header = 0x00
    type = 0x00
    command = 0x00
    data_length = 0x00
    data = bytearray()
    checksum = 0x00
    end = 0x04

    result = bytearray()

    def __init__(self):
        pass

    def encode_checksum(self):
        checksum = self.type ^ self.command ^ self.data_length
        for d in self.data:
            checksum ^= d
        return checksum

class PC2Module(Protocol):
    header = 0x05
    end = 0x04

    def __init__(self):
        super(PC2Module, self).__init__()

    def set_type(self, type):
        self.type = type
        return self

    def set_command(self, command):
        self.command = command
        return self

    def set_data(self, data):
        self.data = bytearray(data)
        self.data_length = len(self.data)
        return self

    def encode(self):
        self.result = bytearray()

        self.checksum = self.encode_checksum()

        self.result.append(self.header)
        self.result.append(self.type)
        self.result.append(self.command)
        self.result.append(self.data_length)
        self.result.extend(self.data)
        self.result.append(self.checksum)
        self.result.append(self.end)

        return self.result


class Module2PC(Protocol):
    end = 0x04

    def __init__(self):
        super(Module2PC, self).__init__()

    def decode(self, byte):
        try:
            byte = bytearray(byte)

            # TODO: Check header success or fail
            self.header = byte[0]
            self.type = byte[1]
            self.command = byte[2]
            self.data_length = byte[3]
            self.data = byte[4:4+self.data_length]
            self.checksum = byte[-2]
            self.end = byte[-1]

            checksum = self.encode_checksum()

            if checksum != self.checksum:
                raise ValueError('Module2PC unmatched checksum')
        except:
            raise ValueError('Module2PC invalid protocol length')

        return self.command, self.data



class BaseLED(object):
    def __init__(self, module):
        self.m = module

        self.type = 0x02
        self.command = {
            'r': 0x00,
            'g': 0x01,
            'b': 0x02,
            'mix': 0x03,
        }
        self.mode = {
            'off': 0x00,
            'on': 0x01,
            'toggle': 0x02,
        }

        self.mix_on_time = 10 # ms


    def __send(self, color, mode, on_delay=0, off_delay=0):
        on_low_bit = on_delay & 0xFF
        on_high_bit = (on_delay >> 8) & 0xFF

        off_low_bit = off_delay & 0xFF
        off_high_bit = (off_delay >> 8) & 0xFF

        data = self.m.p2m.set_type(self.type).set_command(self.command[color]).set_data([self.mode[mode], on_high_bit, on_low_bit, off_high_bit, off_low_bit]).encode()
        self.m.send(data)


    def off(self, color='all'):
        """Turn off the LED light
        Args:
            color (str, optional): Color of the LED light. ``r``, ``g`` or ``b``. Defaults to ``all``.
        """
        if color in ['all', 'r']:
            self.__send('r', 'off')

        if color in ['all', 'g']:
            self.__send('g', 'off')

        if color in ['all', 'b']:
            self.__send('b', 'off')


    def on(self, color='all'):
        """Turn on the LED light.
        Args:
            color (str, optional): Color of the LED light. ``r``, ``g`` or ``b``. Defaults to ``all``.
        """
        if color in ['all', 'r']:
            self.__send('r', 'on')

        if color in ['all', 'g']:
            self.__send('g', 'on')

        if color in ['all', 'b']:
            self.__send('b', 'on')


    def toggle(self, color='all'):
        """Toggle the LED light. Turn off while the light on status and turn on while the light off.
        Args:
            color (str, optional): Color of the LED light. ``r``, ``g`` or ``b``. Defaults to ``all``.
        """
        if color in ['all', 'r']:
            self.__send('r', 'toggle')

        if color in ['all', 'g']:
            self.__send('g', 'toggle')

        if color in ['all', 'b']:
            self.__send('b', 'toggle')


    def mix(self, rgb):
        """Mix the color of base LED light.

        Args:
            rgb ([tuple]): Intensity of LED lights. Range from 0 to 100. e.g) (100, 0, 0) red light
        """
        r, g, b = rgb

        on_low_bit = self.mix_on_time & 0xFF
        on_high_bit = (self.mix_on_time >> 8) & 0xFF

        data = self.m.p2m.set_type(self.type).set_command(self.command['mix']).set_data([on_high_bit, on_low_bit, r, g, b]).encode()
        self.m.send(data)



import setuptools
import sys

if sys.version_info[0] > 2:
    with open("README.md", "r", encoding="utf-8") as fh:
        long_description = fh.read()
else:
    with open("README.md", "r") as fh:
        long_description = fh.read()

setuptools.setup(
    name="DynamiKontrol",
    version="0.1.2",
    author="Taehee Lee",
    author_email="kairess87@gmail.com",
    description="DynamiKontrol Python API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/TheMatrixGroup/DynamiKontrol',
    license='MIT',
    install_requires=['pyserial>=3.5'],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: System :: Hardware",
        "Topic :: System :: Hardware :: Hardware Drivers"
    ],
    packages=setuptools.find_packages(),
    python_requires=">=3.4.2",
)



import time

class Servo(object):
    """Servo motor submodule class.

    .. highlight:: python
    .. code-block:: python

        from dynamikontrol import Module
        import time

        module = Module()

        module.motor.angle(0)
        time.sleep(2)

        while True:
            module.motor.angle(45)
            time.sleep(2)

            module.motor.angle(-45)
            time.sleep(2)

        module.disconnect()

    Args:
        module (object): Module object.
    """
    def __init__(self, module):
        self.m = module

        self.type = 0x03
        self.command = {
            'angle': 0x00,
            'angle_period': 0x01,
            'set_offset': 0x02,
            'get_offset': 0x81
        }


    def angle(self, angle, period=None):
        """Control the angle of motor.

        Args:
            angle (int): If ``angle > 0`` moves along clockwise, otherwise moves along counter clockwise. ``angle`` must be between ``-85`` to ``85`` in degrees.
            period (uint): Control period. ``period`` must be between ``0`` to ``65535`` in millisecond. Defaults to ``None``.
        """
        direction = 0x00 if angle >= 0 else 0x01
        angle_hex = abs(angle)

        if period is None:
            data = self.m.p2m.set_type(self.type).set_command(self.command['angle']).set_data([direction, angle_hex]).encode()
        else:
            if period < 0 or period > 65535:
                raise ValueError('Motor period value must be between 0 to 65535 in millisecond.')

            period_h = (period >> 8) & 0xff
            period_l = period & 0xff

            data = self.m.p2m.set_type(self.type).set_command(self.command['angle_period']).set_data([direction, angle_hex, period_h, period_l]).encode()
        self.m.send(data)


    def get_offset(self):
        """Get offset of the motor.

        Returns:
            float: Offset of the motor in degree.
        """
        data = self.m.p2m.set_type(self.type).set_command(self.command['get_offset']).set_data([]).encode()

        command, received_data = self.m._manual_send_receive(data, 3 + 6)

        direction = 1 if received_data[0] == 0 else -1
        angle_int = received_data[1]
        angle_point = received_data[2] / 10.

        return direction * (angle_int + angle_point)


    def set_offset(self, angle):
        """Set offset of the motor. If the motor angle is inclined slightly even angle set to 0, you can adjust offset of the motor manually.

        Args:
            angle (float): Offset of the motor in degree. e.g) 17.5
        """
        direction = 0x00 if angle >= 0 else 0x01
        angle_hex = abs(angle)
        angle_int = int(angle_hex)
        angle_point = int(round(angle_hex - angle_int, 1) * 10)

        data = self.m.p2m.set_type(self.type).set_command(self.command['set_offset']).set_data([direction, angle_int, angle_point]).encode()
        self.m.send(data)

        time.sleep(0.1)
        self.angle(0)
        time.sleep(0.1)


class Motor(object):
    def __init__(self, module):
        self.m = module

        if self.m.pid == '0001':
            self.motor = Servo(module=self.m)
        elif self.m.pid == '0002':
            raise NotImplementedError('Speed motor module is not implemented yet.')

    def angle(self, *args, **kwargs):
        self.motor.angle(*args, **kwargs)

    def get_offset(self, *args, **kwargs):
        return self.motor.get_offset(*args, **kwargs)

    def set_offset(self, *args, **kwargs):
        self.motor.set_offset(*args, **kwargs)



from dynamikontrol.Module import Module
from dynamikontrol.Protocol import Protocol
from dynamikontrol.Timer import Timer
from dynamikontrol.LED import LED
from dynamikontrol.BaseLED import BaseLED
from dynamikontrol.Motor import Motor
